#
# Cambalache Property wrapper
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
from gi.repository import GObject

from .cmb_base_objects import CmbBaseProperty, CmbBaseObject
from .cmb_property_info import CmbPropertyInfo
from . import utils
from cambalache import _, getLogger

logger = getLogger(__name__)


class CmbProperty(CmbBaseProperty):
    object = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)
    info = GObject.Property(type=CmbPropertyInfo, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, **kwargs):
        self._init = True
        super().__init__(**kwargs)
        self._init = False
        self.version_warning = None

        owner_info = self.project.type_info.get(self.info.owner_id, None)
        self.library_id = owner_info.library_id
        self._update_version_warning()

        self.connect("notify", self.__on_notify)

    def __str__(self):
        return f"CmbProperty<{self.object.type_id} {self.info.owner_id}::{self.property_id}>"

    def __on_notify(self, obj, pspec):
        if self.info.type_id in ["GFile", "GMenuModel"]:
            self._update_version_warning()

        self.object._property_changed(self, pspec.name)

    def __db_get(self, column):
        row = self.project.db.execute(
            f"SELECT {column} FROM object_property WHERE ui_id=? AND object_id=? AND owner_id=? AND property_id=?;",
            (self.ui_id, self.object_id, self.owner_id, self.property_id),
        ).fetchone()
        return row[0] if row is not None else None

    def has_value(self):
        return self.db_get(
            "SELECT count(ui_id) FROM object_property WHERE ui_id=? AND object_id=? AND owner_id=? AND property_id=?;",
            (self.ui_id, self.object_id, self.owner_id, self.property_id),
        ) > 0

    def __db_set(self, **kwargs):
        # Do not use REPLACE INTO, to make sure both INSERT and UPDATE triggers are used

        columns = tuple(kwargs.keys())
        values = tuple(kwargs.values())
        placeholders = ",".join((["?"] * len(values)))

        # Ensure row exists
        if self.has_value():
            # Execute update statement and return row values
            statement = ",".join([f"{col}=?" for col in columns])
            row = self.project.db.execute(
                f"""
                UPDATE object_property SET {statement} WHERE ui_id=? AND object_id=? AND owner_id=? AND property_id=?
                RETURNING
                    value, translatable, comment, translation_context, translation_comments,
                    inline_object_id,
                    bind_source_id, bind_owner_id, bind_property_id, bind_flags,
                    binding_expression_id, binding_expression_object_id;
                """,
                values + (self.ui_id, self.object_id, self.owner_id, self.property_id)
            ).fetchone()

            value, *others = row

            # If value is the same as the default and the rest are none/false we can remove the row
            # Do not reset properties that where loaded from xml with the default value
            if not self.serialize_default_value and \
               (value is None or value == self.info.default_value) and \
               all(not val for val in others):
                self.__reset()
        else:
            self.project.db.execute(
                f"""
                INSERT INTO object_property (ui_id, object_id, owner_id, property_id, {",".join(columns)})
                VALUES (?, ?, ?, ?, {placeholders});
                """,
                (self.ui_id, self.object_id, self.owner_id, self.property_id) + values,
            )

        self.__update_internal_child()

        if self._init is False:
            self.object._property_changed(self, columns[0])

    @GObject.Property(type=str)
    def value(self):
        return self.__db_get("value") or self.info.default_value

    @value.setter
    def _set_value(self, value):
        self.__db_set(value=value)

    @GObject.Property(type=bool, default=False)
    def translatable(self):
        return self.__db_get("translatable")

    @translatable.setter
    def _set_translatable(self, value):
        self.__db_set(translatable=value)

    @GObject.Property(type=str)
    def translation_context(self):
        return self.__db_get("translation_context")

    @translation_context.setter
    def _set_translation_context(self, value):
        self.__db_set(translation_context=value)

    @GObject.Property(type=str)
    def translation_comments(self):
        return self.__db_get("translation_comments")

    @translation_comments.setter
    def _set_translation_comments(self, value):
        self.__db_set(translation_comments=value)

    def __reset(self):
        self.project.db.execute(
            "DELETE FROM object_property WHERE ui_id=? AND object_id=? AND owner_id=? AND property_id=?;",
            (self.ui_id, self.object_id, self.owner_id, self.property_id),
        )

    def reset(self):
        if self.info.internal_child:
            self.project.history_push(_("Unset {obj} {prop} {prop_type}").format(**self.__get_msgs()))

        self.__reset()
        self.__update_internal_child()
        self.notify("value")

        if self.info.internal_child:
            self.project.history_pop()

    def __update_internal_child(self):
        internal_info = self.info.internal_child
        if internal_info and internal_info.internal_parent_id:
            logger.warning("Adding an internal child within an internal child automatically is not implemented")
            return
        elif internal_info is None:
            return

        value = self.value
        child_id = self.db_get(
            "SELECT object_id FROM object WHERE ui_id=? AND parent_id=? AND internal=?",
            (self.ui_id, self.object_id, internal_info.internal_child_id)
        )

        if value and not child_id:
            self.project.add_object(
                self.ui_id,
                internal_info.internal_type,
                parent_id=self.object_id,
                internal=internal_info.internal_child_id
            )
        elif child_id:
            internal_child = self.project.get_object_by_id(self.ui_id, child_id)
            if internal_child:
                self.project.remove_object(internal_child, allow_internal_removal=True)

    def __get_msgs(self, value=None):
        return {
            "obj": self.object.display_name_type,
            "prop": self.property_id,
            "prop_type": _("property"),
            "value": str(value)
        }

    @GObject.Property(type=CmbBaseProperty)
    def bind_property(self):
        bind_source_id = self.bind_source_id
        bind_property_id = self.bind_property_id

        if bind_source_id and bind_property_id:
            source = self.project.get_object_by_id(self.ui_id, bind_source_id) if bind_property_id else None
            return source.properties_dict.get(bind_property_id, None) if source else None

        return None

    @bind_property.setter
    def _set_bind_property(self, bind_property):
        bind_source_id, bind_owner_id, bind_property_id, bind_flags = (None, None, None, None)

        if bind_property:
            obj = bind_property.object

            bind_source_id = obj.object_id
            bind_owner_id = bind_property.owner_id
            bind_property_id = bind_property.property_id
            bind_flags = bind_property.bind_flags

            if obj.ui_id == self.ui_id and obj.object_id == self.bind_source_id and \
               self.bind_property_id == bind_property.property_id:
                return

            self.project.history_push(
                _("Bind {object}::{property} to {bind_object}::{bind_property}").format(
                    object=self.object.display_name_type,
                    property=self.property_id,
                    bind_object=obj.display_name_type,
                    bind_property=bind_property.property_id
                )
            )
        else:
            if not self.bind_source_id and not self.bind_property_id:
                return

            self.project.history_push(
                _("Clear {object}::{property} binding").format(
                    object=self.object.display_name_type,
                    property=self.property_id
                )
            )

        self.__db_set(
            bind_source_id=bind_source_id,
            bind_owner_id=bind_owner_id,
            bind_property_id=bind_property_id,
            bind_flags=bind_flags
        )

        self.project.history_pop()
        self.project._object_property_binding_changed(self.object, self)

    @GObject.Property(type=CmbBaseObject)
    def binding_expression(self):
        return self.project.get_object_by_id(self.ui_id, self.binding_expression_id)

    @binding_expression.setter
    def _set_binding_expression(self, binding_expression):
        if binding_expression:

            if binding_expression.ui_id == self.ui_id and binding_expression.object_id == self.binding_expression_id:
                return

            self.project.history_push(
                _("Bind {object}::{property} to {expression_object}").format(
                    object=self.object.display_name_type,
                    property=self.property_id,
                    expression_object=binding_expression.display_name_type
                )
            )
        else:
            if not self.binding_expression_id:
                return

            self.project.history_push(
                _("Clear {object}::{property} binding expression").format(
                    object=self.object.display_name_type,
                    property=self.property_id
                )
            )

        self.__db_set(binding_expression_id=binding_expression.object_id)

        self.project.history_pop()
        self.project._object_property_binding_changed(self.object, self)

    @GObject.Property(type=CmbBaseObject)
    def binding_expression_object(self):
        return self.project.get_object_by_id(self.ui_id, self.binding_expression_id)

    @binding_expression.setter
    def _set_binding_expression_object(self, expression_object):
        if expression_object:
            if expression_object.ui_id == self.ui_id and expression_object.object_id == self.expression_object_id:
                return

            self.project.history_push(
                _("Bind {object}::{property} to {expression_object}").format(
                    object=self.object.display_name_type,
                    property=self.property_id,
                    expression_object=expression_object.display_name_type
                )
            )
        else:
            if not self.binding_expression_id:
                return

            self.project.history_push(
                _("Clear {object}::{property} binding expression object").format(
                    object=self.object.display_name_type,
                    property=self.property_id
                )
            )

        self.__db_set(binding_expression_id=expression_object.object_id)
        self.project.history_pop()
        self.project._object_property_binding_changed(self.object, self)

    @GObject.Property(type=bool, default=False)
    def serialize_default_value(self):
        return self.__db_get("serialize_default_value")

    @serialize_default_value.setter
    def _set_serialize_default_value(self, value):
        self.__db_set(serialize_default_value=value, value=self.value)

    def _update_version_warning(self):
        warnings = []
        target = self.object.ui.get_target(self.library_id)

        deprecation_warning = utils.get_version_warning(
            target, self.info.version, self.info.deprecated_version, self.property_id
        )
        if deprecation_warning:
            warnings.append(deprecation_warning)

        if self.project.target_tk == "gtk-4.0":
            target = self.object.ui.get_target("gtk")
            if target is not None:
                version = utils.parse_version(target)

                if self.info.type_id == "GFile" and (version is None or utils.version_cmp(version, (4, 16, 0)) < 0) and \
                   (self.value and not os.path.isabs(self.value)):
                    warnings.append(_("Warning: GFile uri needs to be absolute for Gtk < 4.16"))
                if self.info.type_id == "GMenuModel" and (version is None or utils.version_cmp(version, (4, 18, 0)) < 0) and \
                   self.inline_object_id:
                    # FIXME: mark as translatable
                    warnings.append("Inline menus are not supported in Gtk < 4.18")

        self.version_warning = "\n".join(warnings) if len(warnings) else None

    def __clear_expression_inline_object(self):
        binding_expression_id = self.binding_expression_id

        if binding_expression_id:
            expression_source = self.project.get_object_by_id(self.ui_id, binding_expression_id)
            self.project.remove_object(expression_source)

    def clear_binding(self):
        if self.bind_property is None and self.bind_flags is None and self.binding_expression_id is None and \
           self.binding_expression_object_id is None:
            return

        self.project.history_push(
            _("Clear {object}::{property} binding").format(
                object=self.object.display_name_type,
                property=self.property_id
            )
        )

        self.__clear_expression_inline_object()

        bind_source_id = self.bind_source_id != 0
        bind_owner_id = self.bind_owner_id is not None
        bind_property_id = self.bind_property_id is not None
        bind_flags = self.bind_flags is not None
        binding_expression_id = self.binding_expression_id != 0
        binding_expression_object_id = self.binding_expression_object_id != 0

        self.__db_set(
            bind_source_id=None,
            bind_owner_id=None,
            bind_property_id=None,
            bind_flags=None,
            binding_expression_id=None,
            binding_expression_object_id=None
        )

        if bind_source_id:
            self.notify("bind-source-id")
        if bind_owner_id:
            self.notify("bind-owner-id")
        if bind_property_id:
            self.notify("bind-property-id")
        if bind_flags:
            self.notify("bind-flags")
        if binding_expression_id:
            self.notify("binding-expression-id")
        if binding_expression_object_id:
            self.notify("binding-expression-object-id")

        self.project.history_pop()

    def set_binding_expression_type(self, expression_type):
        info = self.project.type_info.get(expression_type)

        if info is None:
            return

        self.project.history_push(
            _("Bind {object}::{property} to {expression_type}").format(
                object=self.object.display_name_type,
                property=self.property_id,
                expression_type=expression_type
            )
        )

        self.__clear_expression_inline_object()

        self.project.add_object(
            self.ui_id,
            info.type_id,
            parent_id=self.object.object_id,
            inline_property=self.property_id,
            inline_binding_expression=True
        )

        self.notify("binding-expression-id")
        self.project.history_pop()




#!/usr/bin/python3
#
# Cambalache UI Maker developer mode
#
# Copyright (C) 2021-2025  Juan Pablo Ugarte
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import os
import gi
import sys
import stat
import signal
import subprocess

basedir = os.path.join(os.path.split(os.path.dirname(__file__))[0])
localdir = os.path.join(basedir, ".local")

cambalachedir = os.path.join(basedir, "cambalache")
localpkgdatadir = os.path.join(localdir, "share", "cambalache")
catalogsdir = os.path.join(localpkgdatadir, "catalogs")
localbindir = os.path.join(localdir, "bin")

repository = gi.Repository.get_default()

# Workaround for distros that say are multiarch but install things in lib/
for locallibdir in [
    os.path.join(localdir, "lib", sys.implementation._multiarch),
    os.path.join(localdir, "lib")
]:
    LD_LIBRARY_PATH = [locallibdir, f"{locallibdir}/cambalache", f"{locallibdir}/cmb_catalog_gen"]
    for path in LD_LIBRARY_PATH:
        repository.prepend_library_path(path)

    GI_TYPELIB_PATH = [f"{locallibdir}/girepository-1.0", f"{locallibdir}/cambalache", f"{locallibdir}/cmb_catalog_gen"]
    for path in GI_TYPELIB_PATH:
        repository.prepend_search_path(path)

    for var, value in [
        ("LD_LIBRARY_PATH", ":".join(LD_LIBRARY_PATH)),
        ("GI_TYPELIB_PATH", ":".join(GI_TYPELIB_PATH)),
        ("PKG_CONFIG_PATH", os.path.join(locallibdir, "pkgconfig")),
        ("GSETTINGS_SCHEMA_DIR", os.path.join(localdir, "share", "glib-2.0", "schemas")),
        ("XDG_DATA_DIRS", os.path.join(localdir, "share")),
        ("PYTHONPATH", os.path.join(localdir, "lib", "python3", "dist-packages"))
    ]:
        if var in os.environ:
            old_value = os.environ[var]
            os.environ[var] = f"{value}:{old_value}"
        else:
            os.environ[var] = value

sys.path.insert(1, basedir)
sys.path.insert(1, cambalachedir)
sys.path.insert(1, localbindir)

from gi.repository import GLib  # noqa: E402


def find_program_in_path(program):
    retval = GLib.find_program_in_path(program)
    if retval is None:
        print(f"Could not find {program} in PATH")
    return retval


signal.signal(signal.SIGINT, signal.SIG_DFL)


def dev_config(filename, content):
    meson_mtime = os.path.getmtime(os.path.join(basedir, "meson.build"))

    abspath = os.path.join(basedir, filename)
    if not os.path.exists(abspath) or meson_mtime > os.path.getmtime(abspath):
        with open(abspath, "w") as fd:
            fd.write(content)


def configure_file(input_file, output_file, config):
    with open(input_file, "r") as fd:
        content = fd.read()

        for key in config:
            content = content.replace(f"@{key}@", config[key])

        with open(output_file, "w") as outfd:
            outfd.write(content)


def get_version():
    meson = open(os.path.join(basedir, "meson.build"))
    version = None
    fileformatversion = None

    for line in meson:
        line = line.strip()
        if version is None and line.startswith("version"):
            tokens = line.split(":")
            version = tokens[1].strip().replace("'", "").replace(",", "")
        elif fileformatversion is None and line.startswith("fileformatversion"):
            tokens = line.split("=")
            fileformatversion = tokens[1].strip().replace("'", "")

    meson.close()

    return (version, fileformatversion)


def run_meson():
    for prog in ["meson", "ninja", "cc", "pkg-config", "g-ir-compiler", "g-ir-scanner"]:
        if GLib.find_program_in_path(prog) is None:
            print(f"{prog} is needed to compile Cambalache private library")
            return -1

    builddir = os.path.join(localdir, "build")
    ninja = os.path.join(builddir, "build.ninja")

    if not os.path.exists(ninja):
        GLib.mkdir_with_parents(builddir, 0o700)
        os.system(f"meson setup --wipe --buildtype=debug --prefix={localdir} {builddir}")

    result = subprocess.run(['ninja', '-C', builddir], stdout=subprocess.PIPE)
    if result.returncode == 0:
        if "ninja: no work to do." not in result.stdout.decode('utf-8'):
            os.system(f"ninja -C {builddir} install")
    else:
        print(result.stdout.decode('utf-8'))

    return result.returncode


def cmb_init_dev():
    version, fileformatversion = get_version()

    # Create config files pointing to source directories
    dev_config(
        os.path.join(cambalachedir, "config.py"),
        f"""VERSION = '{version}'
FILE_FORMAT_VERSION = '{fileformatversion}'
pkgdatadir = '{localpkgdatadir}'
merenguedir = '{cambalachedir}'
catalogsdir = '{catalogsdir}'
privatecambalachedir = '{localdir}'
""",
    )

    # Create config files pointing to source directories
    dev_config(
        os.path.join(cambalachedir, "merengue/config.py"),
        f"""VERSION = '{version}'
pkgdatadir = '{localpkgdatadir}'
merenguedir = '{cambalachedir}'
privatecambalachedir = '{localdir}'
""",
    )

    coverage_bin = GLib.find_program_in_path("python3-coverage")

    # Create merengue bin script
    if coverage_bin and os.getenv("COVERAGE_PROCESS_START", None):
        merengue_shebang = coverage_bin + " run"
    else:
        merengue_shebang = GLib.find_program_in_path("python3")

    configure_file(
        os.path.join(cambalachedir, "merengue", "merengue.in"),
        os.path.join(cambalachedir, "merengue", "merengue"),
        {"PYTHON": merengue_shebang, "merenguedir": cambalachedir},
    )
    os.chmod(os.path.join(cambalachedir, "merengue", "merengue"), stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

    return run_meson()


if __name__ == "__main__":
    exit(cmb_init_dev())



#!/usr/bin/pytest

"""
Test Project Catalogs API
"""

from cambalache import CmbProject


def cmb_catalog_data_test(target_tk):
    # Create new project
    project = CmbProject(target_tk=target_tk)

    # Check for a known iface
    row = project.db.execute("SELECT parent_id FROM type WHERE type_id='GtkOrientable';").fetchall()
    assert row is not None
    assert len(row) == 1
    assert row[0][0] == "interface"

    # Check a know iface property
    row = project.db.execute(
        "SELECT property_id FROM property WHERE owner_id='GtkOrientable' AND property_id='orientation';"
    ).fetchall()
    assert row is not None
    assert len(row) == 1

    # Check a type has the iface
    row = project.db.execute("SELECT iface_id FROM type_iface WHERE type_id='GtkBox' AND iface_id='GtkOrientable';").fetchall()
    assert row is not None
    assert len(row) == 1


def test_gtk3_project_catalogs_data():
    cmb_catalog_data_test("gtk+-3.0")


def test_gtk4_project_catalogs_data():
    cmb_catalog_data_test("gtk-4.0")



#
# CmbGraphicsOffload
#
# Copyright (C) 2026 Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GLib, GObject, Gtk
from cambalache import config


class CmbGraphicsOffload(Gtk.GraphicsOffload, Gtk.Scrollable):
    __gtype_name__ = "CmbGraphicsOffload"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__hadjustment = None
        self.__vadjustment = None

        self.connect("notify::child", self.__on_child_notify)

    @GObject.Property(type=Gtk.Adjustment)
    def hadjustment(self):
        return self.__hadjustment

    @hadjustment.setter
    def _set_hadjustment(self, adjustment):
        self.__hadjustment = adjustment
        if self.props.child:
            self.props.child.props.hadjustment = adjustment

    @GObject.Property(type=Gtk.Adjustment)
    def vadjustment(self):
        return self.__vadjustment

    @vadjustment.setter
    def _set_vadjustment(self, adjustment):
        self.__vadjustment = adjustment
        if self.props.child:
            self.props.child.props.vadjustment = adjustment

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def hscroll_policy(self):
        return self.props.child.props.hscroll_policy if self.props.child else Gtk.ScrollablePolicy.MINIMUM

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def vscroll_policy(self):
        return self.props.child.props.vscroll_policy if self.props.child else Gtk.ScrollablePolicy.MINIMUM

    def __on_child_notify(self, obj, pspec):
        if self.props.child:
            self.props.child.props.hadjustment = self.__hadjustment
            self.props.child.props.vadjustment = self.__vadjustment



#
# CmbObjectDataEditor - Cambalache Object Data Editor
#
# Copyright (C) 2022-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_type_info import CmbTypeDataInfo
from .cmb_object_data import CmbObjectData
from .control import cmb_create_editor
from cambalache import _


# Everyone knows that debugging is twice as hard as writing a program in the first place.
# So if you’re as clever as you can be when you write it, how will you ever debug it?
# -- Brian Kernighan, 1974
#
# TODO: rewrite this!

@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_object_data_editor.ui")
class CmbObjectDataEditor(Gtk.Box):
    __gtype_name__ = "CmbObjectDataEditor"

    info = GObject.Property(type=CmbTypeDataInfo, flags=GObject.ParamFlags.READWRITE)

    label = Gtk.Template.Child()
    top_box = Gtk.Template.Child()
    add_child = Gtk.Template.Child()
    add_only_child = Gtk.Template.Child()
    remove_button = Gtk.Template.Child()
    grid = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.__object = None
        self.__data = None

        self.__size_group = None
        self.__value_editor = None
        self.__arg_editors = {}
        self.__editors = []

        super().__init__(**kwargs)

        self.__update_view()

    @Gtk.Template.Callback("on_add_only_child_clicked")
    def __on_add_only_child_clicked(self, button):
        info = self.data.info if self.data else self.info

        # Do not add a menu if there is only one child type
        child_info = list(info.children.values())[0]
        self.__on_child_button_clicked(button, child_info)

    @Gtk.Template.Callback("on_remove_clicked")
    def __on_remove_clicked(self, button):
        if self.info:
            self.object.remove_data(self.__data)
        elif self.__data:
            self.__data.parent.remove_data(self.__data)

    @GObject.Property(type=GObject.Object)
    def object(self):
        return self.__object

    @object.setter
    def _set_object(self, value):
        if self.__object:
            self.__object.disconnect_by_func(self.__on_data_added)
            self.__object.disconnect_by_func(self.__on_data_removed)

        self.__object = value

        if self.__object:
            self.__object.connect("data-added", self.__on_data_added)
            self.__object.connect("data-removed", self.__on_data_removed)

    @GObject.Property(type=CmbObjectData)
    def data(self):
        return self.__data

    @data.setter
    def _set_data(self, value):
        if self.__data:
            self.__data.disconnect_by_func(self.__on_data_data_added)
            self.__data.disconnect_by_func(self.__on_data_data_removed)
            self.__data.disconnect_by_func(self.__on_data_arg_changed)

        # Clear old editors
        for editor in self.__editors:
            self.grid.remove(editor)
        self.__editors = []

        self.__data = value

        if self.__data:
            self.__data.connect("data-added", self.__on_data_data_added)
            self.__data.connect("data-removed", self.__on_data_data_removed)
            self.__data.connect("arg-changed", self.__on_data_arg_changed)

    def __update_arg(self, key):
        if not self.data:
            return

        editor = self.__arg_editors.get(key, None)
        if editor:
            val = self.data.get_arg(key)

            # Only update if there is a change
            if val != editor.cmb_value:
                editor.cmb_value = val

    def __on_data_data_added(self, parent, data):
        self.__add_data_editor(data)

    def __on_data_data_removed(self, parent, data):
        self.__remove_data_editor(data)

    def __on_data_arg_changed(self, data, key):
        self.__update_arg(key)

    def __on_data_added(self, obj, data):
        if self.info and self.data is None and self.info == data.info:
            self.data = data
            self.__update_view()

    def __on_data_removed(self, obj, data):
        self.__remove_data_editor(data)

    def __ensure_object_data(self, history_message):
        if self.data:
            return False

        self.object.project.history_push(history_message)
        self.data = self.object.add_data(self.info.key)

        if self.__value_editor:
            GObject.Object.bind_property(
                self.data,
                "value",
                self.__value_editor,
                "cmb-value",
                GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
            )
        return True

    def __on_child_button_clicked(self, button, info):
        msg = _("Add {key}").format(key=info.key)
        history_pushed = self.__ensure_object_data(msg)
        self.data.add_data(info.key)
        if history_pushed:
            self.object.project.history_pop()

    def __context_menu_new(self, info):
        popover = Gtk.Popover(position=Gtk.PositionType.BOTTOM)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True, spacing=4, border_width=4)

        # Add children types
        for child in info.children:
            child_info = info.children[child]
            button = Gtk.ModelButton(label=_("Add {key}").format(key=child_info.key), visible=True)
            button.connect("clicked", self.__on_child_button_clicked, child_info)
            box.append(button)

        popover.set_child(box)

        return popover

    def __on_arg_notify(self, obj, pspec, info):
        msg = _("Set {key} to {value}").format(key=info.key, value=obj.cmb_value)
        history_pushed = self.__ensure_object_data(msg)
        self.data.set_arg(info.key, obj.cmb_value)
        if history_pushed:
            self.object.project.history_pop()

    def __add(self, editor, label=None):
        neditors = len(self.__editors)

        if label:
            self.grid.attach(label, 0, neditors, 1, 1)
            self.__size_group.add_widget(label)

        self.grid.attach(editor, 1, neditors, 1, 1)

        self.__editors.append(editor)

    def __add_data_editor(self, data):
        editor = CmbObjectDataEditor(visible=True, hexpand=True, margin_start=16, object=self.object, data=data)
        self.__add(editor)

    def __remove_data_editor(self, data):
        if self.__data == data:
            self.data = None
            return

        for editor in self.__editors:
            if data == editor.data:
                self.grid.remove(editor)
                self.__editors.remove(editor)
                break

    def __update_view(self):
        if self.data is None and self.info is None:
            return

        info = self.data.info if self.data else self.info
        project = self.__object.project

        if info is None:
            return

        nchildren = len(info.children)

        self.remove_button.props.tooltip_text = _("Remove {key}").format(key=info.key)

        # Add a menu if there is more than one child type
        if nchildren > 1:
            self.add_child.props.popover = self.__context_menu_new(info)
            self.add_child.set_visible(True)
        elif nchildren:
            key = list(info.children.keys())[0]
            self.add_only_child.props.tooltip_text = _("Add {key}").format(key=key)
            self.add_only_child.set_visible(True)

        # Item name
        self.label.props.label = info.key

        self.__size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        self.__size_group.add_widget(self.label)

        # Value
        if info.type_id:
            editor = cmb_create_editor(project, info.type_id, data=self.data, parent=self.object)
            self.__value_editor = editor

            if self.data:
                GObject.Object.bind_property(
                    self.data,
                    "value",
                    self.__value_editor,
                    "cmb-value",
                    GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
                )

            self.top_box.append(editor)

        nargs = len(info.args)

        # Arguments
        for arg in info.args:
            arg_info = info.args[arg]

            editor = cmb_create_editor(project, arg_info.type_id, parent=self.object)
            self.__arg_editors[arg_info.key] = editor

            # Initialize value
            self.__update_arg(arg_info.key)

            # Listen for editor value changes and update argument
            editor.connect("notify::cmb-value", self.__on_arg_notify, arg_info)

            # Special case items with one argument and no value (like styles)
            if nargs == 1 and not info.type_id:
                self.label.props.label = f"{info.key} {arg_info.key}"
                self.top_box.append(editor)
            else:
                label = Gtk.Label(visible=True, label=arg_info.key, xalign=1)
                self.__add(editor, label)

        # Current children
        if self.data:
            for child in self.data.children:
                self.__add_data_editor(child)



#!/usr/bin/pytest
#
# GUI Unit Tests
#
# Copyright (C) 2023-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import os
import cairo

from . import utils

basedir = os.path.dirname(__file__)


# Global GtkApplication and CmbWindow
app = None
window = None


def setup_module(module):
    global app, window
    app, window = utils.cmb_create_app()


def teardown_module(module):
    global app, window
    utils.cmb_destroy_app(app, window)
    app = None
    window = None


def window_assert_screenshot(original_basename, target=None, ui_basename=None):
    # Ensure Image directory
    if target:
        imagedir = os.path.join(basedir, "images", target)
    else:
        imagedir = os.path.join(basedir, "images")
    os.makedirs(imagedir, exist_ok=True)

    # Import UI file
    if ui_basename and target:
        path = os.path.join(basedir, target, ui_basename)
        app.import_file(path)

        original_path = os.path.join(imagedir, f"{ui_basename}.png")
    else:
        original_path = os.path.join(imagedir, original_basename)

    # Get Screenshot
    screenshot = utils.window_screenshot(window)

    # Get original image to compare
    if os.path.exists(original_path):
        original = cairo.ImageSurface.create_from_png(original_path)
    else:
        # Write window surface as original if file is not found and return
        screenshot.write_to_png(original_path)
        return

    # NOTE: Saving or Loading the image as PNG slightly changes the pixel values (premultiplication, etc)
    # So in order to avoid any difference we always compare data processed in exactly the same way
    screenshot = utils.image_reload_as_png(screenshot)

    # Get MSE
    ignore_color = (0xFF, 0x00, 0xFF)
    r, g, b, total = utils.mean_squared_error(original, screenshot, ignore_color)

    assert r is not None

    # Save difference image if its not the same
    if total != 0:
        screenshot.write_to_png(os.path.splitext(original_path)[0] + ".screenshot.png")

    # Check its the same image
    # TODO: find a good value
    assert (r, g, b, total) == (0, 0, 0, 0)


def window_activate_action(action_name):
    window.activate_action(action_name, None)
    utils.process_all_pending_gtk_events()


def window_toggle_group_set_active(toggle_group, value):
    group = utils.find_by_buildable_id(window, toggle_group)
    assert group

    group.set_active_name(value)

    utils.process_all_pending_gtk_events()
    return group


def window_button_clicked(button_name):
    button = utils.find_by_buildable_id(window, button_name)
    assert button
    button.emit("clicked")
    utils.process_all_pending_gtk_events()
    return button


def window_entry_set_text(entry_name, text):
    entry = utils.find_by_buildable_id(window, entry_name)
    assert entry
    entry.set_text(text)
    utils.process_all_pending_gtk_events()
    return entry


def window_select_editor_page(editor_name, page):
    editor = utils.find_by_buildable_id(window, editor_name)
    assert editor
    assert editor.stack

    editor.stack.set_visible_child_name(page)
    utils.process_all_pending_gtk_events()


def window_widget_grab_focus(widget):
    widget = utils.find_by_buildable_id(window, widget)
    assert widget
    widget.grab_focus()


def window_add_object(type_id, obj_id, parent_id, ui_id=1):
    project = window.project

    obj = project.add_object(ui_id, type_id, None, parent_id)
    utils.process_all_pending_gtk_events()
    project.set_selection([obj])
    utils.process_all_pending_gtk_events()

    return obj


def select_object(obj_id=None, ui_id=1):
    project = window.project
    obj = project.get_object_by_id(ui_id, obj_id)
    project.set_selection([obj])
    utils.process_all_pending_gtk_events()
    return obj


# TESTS
def _test_new_project(target):
    # New project view
    window_activate_action("win.create_new")
    window_entry_set_text("name_entry", "test_project")

    window_toggle_group_set_active("toolkit_chooser", target)
    # window_widget_grab_focus(toggle)

    window_assert_screenshot("cambalache_new_project.png", target)


def _test_np_create_button(target):
    window_button_clicked("create_button")
    window_assert_screenshot("cambalache_np_create_button.png", target)


def _test_cmb_window_ui_stack_fragment(target):
    ui = window.project.get_object_by_id(1)

    ui.custom_fragment = """<menu id="menubar">
  <submenu>
    <attribute name="label">File</attribute>
  </submenu>
  <submenu>
    <attribute name="label">Edit</attribute>
  </submenu>
  <submenu>
    <attribute name="label">Help</attribute>
  </submenu>
</menu>"""

    window_select_editor_page("ui_editor", "fragment")
    window_assert_screenshot("cambalache_ui_stack_fragment.png", target)


def _test_cmb_window_ui_stack_requires(target):
    ui = window.project.get_object_by_id(1)

    if target == "gtk+-3.0":
        ui.set_library("gtk+", "3.8")
    else:
        ui.set_library("gtk", "4.2")

    select_object(ui_id=1)
    utils.process_all_pending_gtk_events()

    window_select_editor_page("ui_editor", "requires")
    window_assert_screenshot("cambalache_ui_stack_requires.png", target)


def _test_cmb_window_add_window(target):
    window_add_object("GtkWindow", 1, None)
    window_assert_screenshot("cambalache_add_window.png", target)


def _test_cmb_window_object_stack_layout(target):
    window_add_object("GtkGrid", 2, 1)
    window_add_object("GtkButton", 3, 2)
    window_select_editor_page("object_editor", "layout")
    window_assert_screenshot("cambalache_object_stack_layout.png", target)


def _test_cmb_window_object_stack_signals(target):
    button = window.project.get_object_by_id(1, 3)

    button.add_signal("GtkButton", "clicked", "on_button_clicked")

    window_select_editor_page("object_editor", "signals")
    window_assert_screenshot("cambalache_object_stack_signals.png", target)


def _test_cmb_window_object_stack_accessible(target):
    window_select_editor_page("object_editor", "accessible")
    window_assert_screenshot("cambalache_object_stack_accessible.png", target)


def _test_cmb_window_object_stack_fragment(target):
    button = window.project.get_object_by_id(1, 3)
    button.custom_fragment = '<styles><style name="acustomstyle"></styles>'

    window_select_editor_page("object_editor", "fragment")
    window_assert_screenshot("cambalache_object_stack_fragment.png", target)


# Common Tests
def test_cmb_window():
    window_assert_screenshot("cambalache.png")


# Gtk 3
def test_gtk3_new_project():
    _test_new_project("gtk+-3.0")


def test_gtk3_np_create_button():
    _test_np_create_button("gtk+-3.0")


def test_gtk3_cmb_window_add_window():
    _test_cmb_window_add_window("gtk+-3.0")


def test_gtk3_cmb_window_object_stack_layout():
    _test_cmb_window_object_stack_layout("gtk+-3.0")


def test_gtk3_cmb_window_object_stack_signals():
    _test_cmb_window_object_stack_signals("gtk+-3.0")


def test_gtk3_cmb_window_object_stack_accessible():
    _test_cmb_window_object_stack_accessible("gtk+-3.0")


def test_gtk3_cmb_window_object_stack_fragment():
    _test_cmb_window_object_stack_fragment("gtk+-3.0")


def test_gtk3_cmb_window_ui_stack_requires():
    _test_cmb_window_ui_stack_requires("gtk+-3.0")


def test_gtk3_cmb_window_ui_stack_fragment():
    _test_cmb_window_ui_stack_fragment("gtk+-3.0")


# Reset UI to start with the same tests for Gtk 4
def test_cmb_window_close():
    global window

    window_activate_action("win.close")
    window.destroy()
    app.open_project(None)

    windows = app.get_windows()
    assert len(windows)

    window = windows[0]

    window_assert_screenshot("cambalache.png")


# Gtk 4
def test_gtk4_new_project():
    _test_new_project("gtk-4.0")


def test_gtk4_np_create_button():
    _test_np_create_button("gtk-4.0")


def test_gtk4_cmb_window_add_window():
    _test_cmb_window_add_window("gtk-4.0")


def test_gtk4_cmb_window_object_stack_layout():
    _test_cmb_window_object_stack_layout("gtk-4.0")


def test_gtk4_cmb_window_object_stack_signals():
    _test_cmb_window_object_stack_signals("gtk-4.0")


def test_gtk4_cmb_window_object_stack_accessible():
    _test_cmb_window_object_stack_accessible("gtk-4.0")


def test_gtk4_cmb_window_object_stack_fragment():
    _test_cmb_window_object_stack_fragment("gtk-4.0")


def test_gtk4_cmb_window_ui_stack_requires():
    _test_cmb_window_ui_stack_requires("gtk-4.0")


def test_gtk4_cmb_window_ui_stack_fragment():
    _test_cmb_window_ui_stack_fragment("gtk-4.0")



# Cambalache Property Controls
#
# Copyright (C) 2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from .cmb_boolean_undefined import CmbBooleanUndefined
from .cmb_child_type_combo_box import CmbChildTypeComboBox
from .cmb_entry import CmbEntry
from .cmb_file_button import CmbFileButton
from .cmb_file_entry import CmbFileEntry
from .cmb_icon_name_entry import CmbIconNameEntry
from .cmb_pixbuf_entry import CmbPixbufEntry
from .cmb_property_chooser import CmbPropertyChooser
from .cmb_spin_button import CmbSpinButton
from .cmb_text_buffer import CmbTextBuffer
from .cmb_toplevel_chooser import CmbToplevelChooser
from .cmb_color_entry import CmbColorEntry
from .cmb_enum_combo_box import CmbEnumComboBox
from .cmb_flags_entry import CmbFlagsEntry
from .cmb_object_chooser import CmbObjectChooser
from .cmb_object_list_editor import CmbObjectListEditor
from .cmb_source_view import CmbSourceView
from .cmb_switch import CmbSwitch
from .cmb_text_view import CmbTextView
from .cmb_translatable_popover import CmbTranslatablePopover
from .cmb_suggestion_entry import CmbSuggestionEntry
from .cmb_create_editor import cmb_create_editor



import os
import time
import inspect
import sqlite3


class CmbProfileConnection(sqlite3.Connection):
    def __init__(self, path, **kwargs):
        super().__init__(path, **kwargs)

        self.executescript(
            """
            CREATE TABLE IF NOT EXISTS __profile__ (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                query TEXT NOT NULL,
                plan TEXT,
                executions INTEGER NOT NULL DEFAULT 1,
                total_time INTEGER NOT NULL DEFAULT 0,
                average_time INTEGER NOT NULL DEFAULT 0,
                min_time INTEGER NOT NULL DEFAULT 0,
                max_time INTEGER NOT NULL DEFAULT 0,
                callers JSONB
            );
            """
        )

        # Striped querys PK dictionary
        self._querys = {}

        # Populate querys
        for row in super().execute("SELECT id, query FROM __profile__;"):
            id, query = row
            self._querys[query] = id

    def cursor(self):
        return super(CmbProfileConnection, self).cursor(CmbProfileCursor)

    def execute(self, *args):
        start = time.monotonic_ns()
        retval = super().execute(*args)
        self.log_query(time.monotonic_ns() - start, *args)
        return retval

    def log_query(self, exec_time, *args):
        query = args[0].strip()

        if query.startswith("CREATE") or query.startswith("PRAGMA"):
            return

        caller = inspect.getframeinfo(inspect.stack()[2][0])
        file = os.path.basename(caller.filename).removesuffix('.py')
        function = caller.function
        # Use a different dot to avoid json syntax error
        caller_id = f"{file}․{function}:{caller.lineno}"

        if file == "cmb_db" and function == "execute":
            caller = inspect.getframeinfo(inspect.stack()[3][0])
            file = os.path.basename(caller.filename).removesuffix('.py')
            caller_id = f"{file}․{caller.function}:{caller.lineno} {caller_id}"

        # Convert from nano seconds to micro seconds
        exec_time = int(exec_time / 1000)
        pk_id = self._querys.get(query, None)

        if pk_id is None:
            # Get query plan
            if len(args) > 1:
                c = super().execute(f"EXPLAIN QUERY PLAN {query}", args[1])
            else:
                c = super().execute(f"EXPLAIN QUERY PLAN {query}")

            # Convert plan to a string
            plan = []
            for row in c:
                plan.append(" ".join(str(row)))
            plan = "\n".join(plan)

            # Create new query entry in profile table
            c = super().execute(
                """
                INSERT INTO __profile__(query, plan, total_time, average_time, min_time, max_time, callers)
                VALUES(?, ?, ?, ?, ?, ?, json(?))
                RETURNING id;
                """,
                (query, plan, exec_time, exec_time, exec_time, exec_time, f"""{{"{caller_id}": 1}}""")
            )
            pk_id = c.fetchone()[0]
            self._querys[query] = pk_id
        else:
            # Increment number of executions of this query
            super().execute(
                f"""
                UPDATE __profile__
                SET
                    executions=executions+1,
                    total_time=total_time+?,
                    average_time=total_time/executions,
                    min_time=min(min_time, ?),
                    max_time=max(max_time, ?),
                    callers=json_set(callers, '$.{caller_id}', callers->'$.{caller_id}' + 1)
                WHERE id=?;
                """,
                (exec_time, exec_time, exec_time, pk_id)
            )


class CmbProfileCursor(sqlite3.Cursor):
    def execute(self, *args):
        start = time.monotonic_ns()
        retval = super().execute(*args)
        self.connection.log_query(time.monotonic_ns() - start, *args)
        return retval



#!/bin/python3
#
# update-supporters - Cambalache supporters list tool
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import sys
import csv


header_text = """# Cambalache supporters

Many thanks to all the people that support the project

"""


def get_supporters(filenames):
    retval = []

    for filename in filenames:
        print("Processing", filename)
        with open(filename, newline="") as f:
            reader = csv.DictReader(f)
            for row in reader:
                # Try different columns, patreon and liberapay formats
                for col in ["Name", "patron_public_name", "patron_username", "patron_id"]:
                    name = row.get(col, None)
                    if name is not None and len(name) > 0:
                        break

                total_str = None

                # Get total donations
                for col in ["Lifetime Amount", "sum_received"]:
                    total_str = row.get(col, None)
                    if total_str is not None:
                        break

                total = float(total_str) if total_str is not None else 0
                if total > 0:
                    retval.append((name, total))

    return [x[0] for x in sorted(retval, key=lambda v: v[1], reverse=True)]


def save_supporters(fd, supporters, prefix):
    for name in supporters:
        fd.write(f"{prefix} {name}\n")


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print(f"Usage: {sys.argv[0]} output.md input.csv input2.csv")
        exit()

    supporters = get_supporters(sys.argv[2:])

    with open(sys.argv[1], "w") as fd:
        fd.write(header_text)
        save_supporters(fd, supporters, " - ")
        fd.close()



#!/usr/bin/pytest

"""
Test project export xml for Merengue
"""
import os
from cambalache import CmbProject
from gi.repository import GObject, Gtk
from lxml import etree


basedir = os.path.dirname(__file__)


def xml_get_node(xml, node_query):
    assert xml is not None
    assert node_query is not None

    # Parse xml
    root = etree.fromstring(xml.encode())

    # Find node
    nodes = root.xpath(node_query)
    return nodes[0] if nodes and len(nodes) else None


def xml_check_object_ws_properties(xml, object_query, ws_properties=[]):
    obj_node = xml_get_node(xml, object_query)
    assert obj_node is not None

    for property_id, value, exp in ws_properties:
        # Look for a property node with name==property_id
        prop = obj_node.find(f'property[@name="{property_id}"]')
        assert prop is not None

        # Check value
        if value:
            assert prop.text == value

        # Check property xpath expression
        if exp:
            assert len(prop.xpath(exp)) > 0


def cmb_workspace_default_test(target_tk, type_class, ws_properties):
    """
    Check type_class has ws_properties by default
    """
    project = CmbProject(target_tk=target_tk)

    # Add an UI
    ui = project.add_ui("workspace_defaults.ui")

    # Create object with workspace defaults
    obj = project.add_object(ui.ui_id, type_class)
    assert obj is not None

    # Export UI file in merengue mode (Workspace)
    str_exported = project.db.tostring(ui.ui_id, merengue=True)

    # Check export worked
    assert str_exported is not None

    xml_check_object_ws_properties(str_exported, f"object[@class='{type_class}']", ws_properties)


def test_object_id():
    """
    Make sure merengue output has object ids in the form __cmb__{ui_id}.{object_id}
    """
    project = CmbProject(target_tk="gtk-4.0")

    # Add an UI
    ui = project.add_ui("test.ui")

    objects = []

    # GtkWindow
    win = project.add_object(ui.ui_id, "GtkWindow", "window")
    assert win is not None
    objects.append(win)

    # GtkBox
    box = project.add_object(ui.ui_id, "GtkBox", "box", parent_id=win.object_id)
    assert box is not None
    objects.append(box)

    # GtkButton
    button = project.add_object(ui.ui_id, "GtkButton", "button", parent_id=box.object_id)
    assert button is not None
    objects.append(button)

    # Export UI file in merengue mode (Workspace)
    str_exported = project.db.tostring(ui.ui_id, merengue=True)

    # Check export worked
    assert str_exported is not None

    builder = Gtk.Builder()
    builder.add_from_string(str_exported)

    for obj in objects:
        gobject = builder.get_object(f"__cmb__{ui.ui_id}.{obj.object_id}")
        assert gobject is not None
        assert GObject.type_name(type(gobject)) == obj.type_id


def test_template():
    """
    Make sure merengue output has object ids in the form __cmb__{ui_id}.{object_id}
    """
    project = CmbProject(target_tk="gtk-4.0")

    # Add an UI
    ui = project.add_ui("template.ui")

    # GtkWindow
    win = project.add_object(ui.ui_id, "GtkWindow", "window")
    assert win is not None

    # GtkButton
    button = project.add_object(ui.ui_id, "GtkButton", "button", parent_id=win.object_id)
    assert button is not None

    # Make window a template
    win.name = "MyWindow"
    ui.template_id = win.object_id

    # Export UI file
    str_exported = project.db.tostring(ui.ui_id, merengue=False)
    assert xml_get_node(str_exported, "template[@class='MyWindow' and @parent='GtkWindow']") is not None
    assert xml_get_node(str_exported, "template/child/object[@class='GtkButton' and @id='button']") is not None

    # Export UI file in merengue mode (Workspace)
    str_exported = project.db.tostring(ui.ui_id, merengue=True)
    assert xml_get_node(str_exported, "object[@class='GtkWindow' and @id='__cmb__1.1']") is not None
    assert xml_get_node(str_exported, "object/child/object[@class='GtkButton' and @id='__cmb__1.2']") is not None

    ui2 = project.add_ui("test.ui")
    mywin = project.add_object(ui2.ui_id, "MyWindow", "window")
    assert mywin is not None

    # Export UI file
    str_exported = project.db.tostring(ui2.ui_id, merengue=False)
    assert xml_get_node(str_exported, "object[@class='MyWindow' and @id='window']") is not None

    # Export UI file in merengue mode (Workspace)
    str_exported = project.db.tostring(ui2.ui_id, merengue=True)
    assert xml_get_node(str_exported, "object[@class='GtkWindow']") is not None
    assert xml_get_node(str_exported, "object/child/object[@class='GtkButton']") is not None


def test_no_signals():
    """
    Make sure merengue output does not have signals declaration to avoid GtkBuilder errors not finding the callbacks
    """
    path = os.path.join(basedir, "gtk+-3.0", "signals.ui")

    project = CmbProject(target_tk="gtk+-3.0")
    ui, msgs, detail_msg = project.import_file(path)
    str_exported = project.db.tostring(ui.ui_id, merengue=True)

    root = etree.fromstring(str_exported.encode())

    assert root.find("signal") is None


def test_gtk4_stack_page_workspace_default():
    cmb_workspace_default_test(
        "gtk-4.0",
        "GtkStackPage",
        [("child", None, "object[@class='GtkLabel']/property[@name='label' and text()='Empty Page']")],
    )


def test_gtk4_template_inline_object():
    project = CmbProject(filename=os.path.join(basedir, "gtk-4.0", "template_inline_object.cmb"))

    # Export UI file in merengue mode (Workspace)
    str_exported = project.db.tostring(1, merengue=True)

    # Check export worked
    assert str_exported is not None

    # Parse xml
    root = etree.fromstring(str_exported.encode())

    # Find object node
    label = root.find("object[@class='GtkWindow']/property/object[@class='GtkBox']/child/object[@class='GtkLabel']")
    assert label is not None

    label_property = label.find("property[@name='label']")
    assert label_property is not None

    assert label_property.text == "a label inside an inline object"



#
# Cambalache Tutor
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#
# Based on glade-intro.c (C) 2017-2018 Juan Pablo Ugarte
#

from gi.repository import GObject, GLib, Gdk, Gtk
from enum import Enum
from collections import namedtuple
from cambalache import utils


class CmbTutorState(Enum):
    NULL = 1
    PLAYING = 2
    PAUSED = 3


class CmbTutorPosition(Enum):
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3
    CENTER = 4


ScriptNode = namedtuple("ScriptNode", "widget text delay name position")


class CmbTutor(GObject.GObject):
    __gsignals__ = {
        "show-node": (GObject.SignalFlags.RUN_LAST, None, (str, Gtk.Widget)),
        "hide-node": (GObject.SignalFlags.RUN_LAST, None, (str, Gtk.Widget)),
    }

    window = GObject.Property(type=Gtk.Window, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, script, **kwargs):
        # List of ScriptNode
        self.script = []

        # Popover to show the script text
        self.popover = None

        # Timeout id for running the script
        self.timeout_id = None

        # Current script node index
        self.current = None

        self.hiding_node = None

        super().__init__(**kwargs)

        for node in script:
            self.__add(*node)

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def state(self):
        if self.timeout_id:
            return CmbTutorState.PLAYING
        elif self.current:
            return CmbTutorState.PAUSED

        return CmbTutorState.NULL

    def __add(self, text, widget_name, delay, name=None, position=CmbTutorPosition.BOTTOM):
        def find_by_css_name_or_buildable_id(widget, name):
            retval = None
            css_name = widget.get_name()

            # Get css name first
            if css_name and css_name != GObject.type_name(widget) and css_name == name:
                return widget

            # then GtkBuildable name
            if isinstance(widget, Gtk.Buildable) and Gtk.Buildable.get_buildable_id(widget) == name:
                return widget

            # or ModelButton name
            if GObject.type_name(widget) == "GtkModelButton" and widget.props.text == name:
                return widget

            for child in utils.widget_get_children(widget):
                retval = find_by_css_name_or_buildable_id(child, name)
                if retval:
                    return retval

            return retval

        widget = find_by_css_name_or_buildable_id(self.window, widget_name)

        if widget:
            self.script.append(ScriptNode(widget, text, delay, name, position))

    def play(self):
        if len(self.script) == 0:
            return

        if self.current is None:
            self.current = 0

        self.__script_play()

        self.notify("state")

    def pause(self):
        if self.timeout_id:
            GLib.source_remove(self.timeout_id)

        self.timeout_id = None
        self.__hide_node(self.current)
        self.notify("state")

    def stop(self):
        self.pause()
        self.current = None
        self.notify("state")

    def __popover_new(self, text):
        popover = Gtk.Popover(autohide=False)
        popover.add_css_class("cmb-tutor")
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, hexpand=True)

        box.append(Gtk.Image(icon_name="dialog-information-symbolic"))
        box.append(Gtk.Label(label=text, vexpand=False, hexpand=True, wrap=True, max_width_chars=24))
        popover.set_child(box)

        return popover

    def __script_transition(self):
        self.timeout_id = GLib.timeout_add(250, self.__script_play)

        self.__hide_current_node()

        # Set next node
        if self.current is not None:
            self.current = self.current + 1 if self.current < (len(self.script) - 1) else None

        return GLib.SOURCE_REMOVE

    def __hide_node(self, index):
        if self.popover:
            self.popover.popdown()
            self.popover = None

        if index is not None:
            node = self.script[index]

            if node.widget:
                node.widget.get_style_context().remove_class("cmb-tutor-highlight")

    def __hide_current_node(self):
        if self.hiding_node:
            return

        self.hiding_node = True

        self.__hide_node(self.current)

        if self.current is not None:
            node = self.script[self.current]
            self.emit("hide-node", node.name, node.widget)

        self.hiding_node = False

    def __script_play(self):
        self.timeout_id = None

        if self.current is None:
            return GLib.SOURCE_REMOVE

        node = self.script[self.current]

        if node and node.text:
            # Ensure the widget is visible
            if not node.widget.is_visible():
                # if the widget is inside a popover pop it up
                parent = node.widget.get_ancestor(Gtk.Popover)
                if parent:
                    parent.popup()

            node.widget.add_css_class("cmb-tutor-highlight")

            # Create popover
            self.popover = self.__popover_new(node.text)
            self.popover.set_parent(node.widget)

            if node.position == CmbTutorPosition.BOTTOM:
                self.popover.set_position(Gtk.PositionType.BOTTOM)
            elif node.position == CmbTutorPosition.LEFT:
                self.popover.set_position(Gtk.PositionType.LEFT)
            elif node.position == CmbTutorPosition.RIGHT:
                self.popover.set_position(Gtk.PositionType.RIGHT)
            elif node.position == CmbTutorPosition.CENTER:
                rect = Gdk.Rectangle()
                rect.x = node.widget.get_allocated_width() / 2
                rect.y = node.widget.get_allocated_height() / 2
                self.popover.set_pointing_to(rect)
                self.popover.set_position(Gtk.PositionType.TOP)

        self.emit("show-node", node.name, node.widget)

        if self.popover:
            self.popover.set_sensitive(True)
            self.popover.popup()

        self.timeout_id = GLib.timeout_add(node.delay * 1000, self.__script_transition)

        return GLib.SOURCE_REMOVE



#!/bin/python3
#
# run-dev - Script to run Cambalache from sources
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import os
import sys
import locale
import gettext

from tools.cmb_init_dev import cmb_init_dev

# Compile deps and install things in .local
cmb_init_dev()

basedir = os.path.join(os.path.split(os.path.dirname(__file__))[0])

# Setup gettext for GtkBuilder
locale.bindtextdomain("cambalache", os.path.join(basedir, ".local", "share", "locale"))
locale.textdomain("cambalache")

# Setup for python code
gettext.bindtextdomain("cambalache", os.path.join(basedir, ".local", "share", "locale"))
gettext.textdomain("cambalache")

from cambalache.app import CmbApplication  # noqa E402

CmbApplication().run(sys.argv)
